from revealer.REVEALER_Cython import runREVEALER
#from REVEALER.CheckGrid import runCheckGrid
from revealer.REVEALER_runbenchmark import runBenchmark
import sys
import pandas as pd
import argparse


def main():
    parser = argparse.ArgumentParser()
    optional = parser._action_groups.pop()
    required = parser.add_argument_group('required arguments')
    parser._action_groups.append(optional)
    required.add_argument("-tf", "--target_file", dest="target_file",
                        help="Name of target file, has to be gct file.", metavar="TARGET_FILE",required=True)
    required.add_argument("-f", "--feature_files", dest="feature_files",
                        help="Name of feature file, has to be gct file.", metavar="FEATURE_FILE",required=True)

    optional.add_argument("-sf", "--seed_files", dest="seed_files",
                        help="Name of seed file.", metavar="SEED_FILES")
    optional.add_argument("-p", "--prefix", dest="prefix",
                        help="Prefix for results files. Default is REVEALER.", metavar="PREFIX")
    optional.add_argument("-sn", "--seed_name", dest="seed_name",
                        help="Seed names to take from seed file or feature file. If multiple exist, separate by ','. Passing text file(ending with 'txt') with column of seed names also works", 
                        metavar="CLASS_FILE")
    optional.add_argument("-g", "--grid", dest="grid",
                        help="Grid size to put kernel on. has to be int. Default is 34.", metavar="GRID")
    optional.add_argument("-t", "--target_name", dest="target_name",
                        help="Name of target in target file. Can be row name or row index(0 based). Default is first row.", metavar="TARGET_NAME")
    optional.add_argument("-k", "--k_size", dest="k",
                        help="Size of kernel indicating k variance far from middle. Has to be int. Default is 5", metavar="K")
    optional.add_argument("-bm", "--bandwidth_multiplication", dest="bandwidth_mult",
                        help="Value of bandwidth multiplier. Has to be float or int. Default is 0.65.", metavar="BANDWIDTH_MULT")
    optional.add_argument("-ba", "--bandwidth_adjustment", dest="bandwidth_adj",
                        help="Value of bandwidth adjustment. Has to be float or int. Default is -0.95", metavar="BANDWIDTH_ADJ")
    optional.add_argument("-d", "--direction", dest="direction",
                        help="Direction of phenotype the features should match. Default is positive.", metavar="DIRECTION")
    optional.add_argument("-m", "--mode", dest="mode",
                        help="Mode to run REVEALER. Can be single or multiple. If set to multiple, format as sample input is required. Default is single.", metavar="MODE")
    optional.add_argument("-ps", "--parameter_set", dest="parameter_set",
                        help="set of parameter. Required if mode is multiple. Check avaiable format in documentation.", metavar="MODE")
    optional.add_argument("-nt", "--num_top", dest="num_top",
                        help="Number of top features picked for intermediate report. Has to be int. Default is 30.", metavar="NUM_TOP")
    optional.add_argument("-lt", "--low_threshold", dest="low_threshold",
                        help="Lower threshold to remove feature. If int passed, threshold of absolute value is made. If float passsed, threshold made by ratio. Default is 3.", metavar="LOW_THRESHOLD")
    optional.add_argument("-ht", "--high_threshold", dest="high_threshold",
                        help="Higher threshold to remove feature. If int passed, threshold of absolute value is made. If float passsed, threshold made by ratio. Default is 0.2.", metavar="HIGH_THRESHOLD")
    optional.add_argument("-ic", "--if_collapse", dest="collapse",
                        help="Indicate if features are collapsed for intermediate report. Has to be True or False. Default is False.", metavar="IF_COLLAPSE")
    optional.add_argument("-ff", "--figure_format", dest="figure_format",
                        help="Format for figure. Can be any figure format available in matplotlib. Default is pdf.", metavar="FIGURE_FORMAT")
    optional.add_argument("-tn", "--thread_number", dest="thread_number",
                        help="Number of thread used to run the program. Default is 1, -1 indicate all thread used.", metavar="THREAD_NUMBER")
    optional.add_argument("-n", "--normalize", dest="normalize",
                        help="Way to normalize the target. Can be standard or zerobase. Default is zerobase.", metavar="NORMALIZE")
    optional.add_argument("-gl", "--gene_locus", dest="gene_locus",
                        help="File name to indicate gene locus. Check documentation for format. Default is None.", metavar="GENE_LOCUS")
    optional.add_argument("-v", "--verbose", dest="verbose",
                        help="Verbose level for report. Setting 0 to get no report. Default is 1.", metavar="VERBOSE")
    optional.add_argument("-mi", "--max_iteration", dest="max_iteration",
                        help="Number of iteration. Has to be int. -1 indicates stop by automatic detection. Default is -1", metavar="MAX_ITERATION")
    optional.add_argument("-s", "--subset", dest="subset",
                        help="Subset of samples to be chosen. Has to be file with one column with subset of sample names in target file. Default is use all columns.", metavar="SUBSET")
    optional.add_argument("-ip", "--if_pvalue", dest="if_pval", 
                        help="Indicate if pvalues are calculated and plotted on figure. Has to be True or Flase. Default is True.", metavar="IF_PVALUE")
    optional.add_argument("-ib", "--if_bootstrap", dest="if_bootstrap", 
                        help="Indicate if variance are calculated by bootstrap and plotted on figure. Has to be True or Flase. Default is True.", metavar="IF_BOOTSTRAP")
    optional.add_argument("-icl", "--if_cluster", dest="if_cluster", 
                        help="Indicate if features are clustered using NMF for intermediate report. Has to be True or Flase. Default is False.", metavar="IF_CLUSTER")
    optional.add_argument("-ii", "--if_intermediate", dest="if_intermediate", 
                        help="Indicate if intermediate reportes are generated. Has to be True or Flase. Default is True.", metavar="IF_INTERMEDIATE")
    optional.add_argument("-o", "--out_folder", dest="out_folder",
                        help="Path to directory to put output files. default is current directory.", metavar="OUT_FOLDER")
    optional.add_argument("-sep", "--separator", dest="separator",
                        help="Separator between gene name and later part.", metavar="SEPARATOR")
    optional.add_argument("-gs", "--gene_set", dest="gene_set",
                        help="Gene Set that is extracted.", metavar="GENE_SET")
    optional.add_argument("-gmt", "--gmt_file", dest="gmt_file",
                        help="gmt file for allele information.", metavar="GMT_FILE")
    optional.add_argument("-a", "--alpha", dest="alpha",
                        help="Power to raise value. Default is 1.", metavar="ALPHA")
    optional.add_argument("-gc", "--grid_check", dest="grid_check",
                        help="Indicate if run Grid Check mode. Default is False.", metavar="GRID_CHECK")
    optional.add_argument("-tif", "--tissue_file", dest="tissue_file",
                        help="File of tissue information.", metavar="TISSUE_FILE")
    optional.add_argument("-gz", "--gzip", dest="gzip",
                        help="If output file should be gzipped.", metavar="GZIP")
    optional.add_argument("-nh", "--neighborhood", dest="neighborhood",
                        help="number of neighborhood.", metavar="NEIGHBORHOOD")
    optional.add_argument("-fr", "--feature_remove", dest="feature_remove",
                        help="features to remove manually.", metavar="FEATURE_REMOVE")
    optional.add_argument("-c", "--color", dest="color",
                        help="tune of color pallette.", metavar="COLOR")
    optional.add_argument("-lw", "--line_width", dest="line_width",
                        help="width of line.", metavar="LINE_WIDTH")
    optional.add_argument("-fs", "--feature_set", dest="feature_set",
                        help="set of feature to use.", metavar="FEATURE_SET")

    args = parser.parse_args()
    args = vars(args)
    print(args)

    feature_files = args['feature_files'].split(',')
    for feature_file in feature_files:
        if feature_file[-3:] != 'gct':
            print('Feature files has to be in gct format.')
            sys.exit(1)
    
    if args['target_file'][-3:] != 'gct':
        print('Target file has to be in gct format.')
        sys.exit(1)

    if args['gmt_file'] != None:
        if args['gmt_file'][-3:] != 'gmt':
            print('gmt file has to be in gmt format.')
            sys.exit(1)
        else:
            gmt_file = args['gmt_file']
    else:
        gmt_file = None

    if args['seed_files'] != None:
        if args['seed_files'][-3:] != 'gct':
            print('Seed file has to be in gct format.')
            sys.exit(1)
        else:
            seed_files = args['seed_files'].split(',')
    else:
        seed_files = args['feature_files'].split(',')

    if args['prefix'] != None:
        prefix = args['prefix']
    else:
        prefix = 'REVEALER'

    if args['seed_name'] != None:
        if args['seed_name'][-3:] != 'txt':
            seed_name =  args['seed_name'].split(',')
        else:
            with open(args['seed_name'],'r') as f:
                seed_name = [line.rstrip() for line in f]
    else:
        seed_name = None

    if args['grid'] != None:
        try:
            grid = int(args['grid'])
        except ValueError:
            print('Grid has to be int')
            sys.exit(1)
    else:
        grid = 34

    if args['target_name'] != None:
        try:
            target_name = int(args['target_name'])
        except ValueError:
            target_name = args['target_name']
    else:
        target_name = 0

    if args['k'] != None:
        try:
            k = int(args['k'])
        except ValueError:
            print('K has to be int')
            sys.exit(1)
    else:
        k = 5

    if args['bandwidth_mult'] != None:
        try:
            bandwidth_mult = float(args['bandwidth_mult'])
        except ValueError:
            print('Bnadwidth multiplier has to be int or float')
            sys.exit(1)
    else:
        bandwidth_mult = 0.65

    if args['bandwidth_adj'] != None:
        try:
            bandwidth_adj = float(args['bandwidth_adj'])
        except ValueError:
            print('Bnadwidth adjustment has to be int or float')
            sys.exit(1)
    else:
        bandwidth_adj = -0.95

    if args['direction'] != None:
        if args['direction'] == 'positive':
            direction = 'pos'
        elif args['direction'] == 'negative':
            direction = 'neg'
        else:
            print('Direction has to be either positive or negative')
            sys.exit(1)
    else:
        direction = 'pos'

    if args['mode'] != None:
        if args['mode'] == 'single':
            mode = 'single'
        elif args['mode'] == 'multiple':
            mode = 'multiple'
            if args['parameter_set'] == None:
                print('Please specify parameter set if use mode multiple.')
                sys.exit(1)
            else:
                if args['parameter_set'][-3:] == 'txt':
                    params = pd.read_csv(args['parameter_set'],sep='\t')
                    prefix = params['prefix'].tolist()
                    grid = params['grid'].tolist()
                    k = params['k'].tolist()
                    bandwidth_mult = params['bandwidth_mult'].tolist()
                    bandwidth_adj = params['bandwidth_adj'].tolist()
                else:
                    paramall = args['parameter_set'].split(';')
                    paramlist = []
                    for i in paramall:
                        paramlist.append(i.split(','))
                    params = pd.DataFrame(paramlist,columns=['prefix','grid','k','bandwidth_mult','bandwidth_adj'])
                    prefix = params['prefix'].tolist()
                    grid = params['grid'].tolist()
                    k = params['k'].tolist()
                    bandwidth_mult = params['bandwidth_mult'].tolist()
                    bandwidth_adj = params['bandwidth_adj'].tolist()
        else:
            print('Mode has to be either single or multiple')
            sys.exit(1)
    else:
        mode = 'single'

    if args['num_top'] != None:
        try:
            num_top = int(args['num_top'])
        except ValueError:
            print('Num_top has to be int')
            sys.exit(1)
    else:
        num_top = 30

    if args['low_threshold'] != None:
        try:
            low_threshold = float(args['low_threshold'])
        except ValueError:
            print('Low_threshold has to be int or float')
            sys.exit(1)
    else:
        low_threshold = 3

    if args['high_threshold'] != None:
        try:
            high_threshold = float(args['high_threshold'])
        except ValueError:
            print('High_threshold has to be int or float')
            sys.exit(1)
    else:
        high_threshold = 100

    if args['collapse'] == None:
        collapse = True
    elif args['collapse'] == 'True':
        collapse = True
    elif args['collapse'] == 'False':
        collapse = False
    else:
        print('Only True or False accepted for collapse')
        sys.exit(1)

    if args['normalize'] == None:
        normalize = 'zerobase'
    elif args['normalize'] == 'standard':
        normalize = 'standard'
    elif args['normalize'] == 'zerobase':
        normalize = 'zerobase'
    else:
        print('Only standard or zerobase accepted for normalize')
        sys.exit(1)

    if args['gene_locus'] == None:
        gene_locus = 'None'
    else:
        gene_locus = args['gene_locus']

    if args['verbose'] != None:
        try:
            verbose = int(args['verbose'])
        except ValueError:
            print('verbose has to be int')
            sys.exit(1)
    else:
        verbose = 1

    if args['max_iteration'] != None:
        try:
            max_iteration = int(args['max_iteration'])
        except ValueError:
            print('max_iteration has to be int')
            sys.exit(1)
    else:
        max_iteration = -1

    if args['thread_number'] != None:
        try:
            thread_number = int(args['thread_number'])
        except ValueError:
            print('thread_number has to be int')
            sys.exit(1)
    else:
        thread_number = 1

    if args['figure_format'] != None:
        figure_format = args['figure_format']
    else:
        figure_format = 'pdf'

    if args['subset'] != None:
        with open(args['subset'],'r') as f:
            subset = [line.rstrip() for line in f]
    else:
        subset = 'no'

    if args['if_pval'] == None:
        if_pval = True
    elif args['if_pval'] == 'True':
        if_pval = True
    elif args['if_pval'] == 'False':
        if_pval = False
    else:
        print('Only True or False accepted for if_pvalue')
        sys.exit(1)

    if args['if_bootstrap'] == None:
        if_bootstrap = True
    elif args['if_bootstrap'] == 'True':
        if_bootstrap = True
    elif args['if_bootstrap'] == 'False':
        if_bootstrap = False
    else:
        print('Only True or False accepted for if_bootstrap')
        sys.exit(1)

    if args['if_cluster'] == None:
        if_cluster = True
    elif args['if_cluster'] == 'True':
        if_cluster = True
    elif args['if_cluster'] == 'False':
        if_cluster = False
    else:
        print('Only True or False accepted for if_cluster')
        sys.exit(1)

    if args['if_intermediate'] == None:
        if_intermediate = True
    elif args['if_intermediate'] == 'True':
        if_intermediate = True
    elif args['if_intermediate'] == 'False':
        if_intermediate = False
    else:
        print('Only True or False accepted for if_intermediate')
        sys.exit(1)

    if args['out_folder'] != None:
        out_folder = args['out_folder']
    else:
        out_folder = './'

    if args['separator'] != None:
        separator = args['separator']
    else:
        separator = '_'

    if args['gene_set'] != None:
        if args['gene_set'][-3:] != 'txt':
            gene_set =  args['gene_set'].split(',')
        else:
            with open(args['gene_set'],'r') as f:
                gene_set = [line.rstrip() for line in f]
    else:
        gene_set = None


    if args['alpha'] != None:
        try:
            alpha = float(args['alpha'])
        except ValueError:
            print('alpha has to be int or float')
            sys.exit(1)
    else:
        alpha = 1

    if args['grid_check'] == None:
        grid_check = False
    elif args['grid_check'] == 'True':
        grid_check = True
    elif args['grid_check'] == 'False':
        grid_check = False
    else:
        print('Only True or False accepted for grid_check')
        sys.exit(1)

    if args['tissue_file'] != None:
        if args['tissue_file'][-3:] != 'gct':
            print('tissue file has to be in gct format.')
            sys.exit(1)
        else:
            tissue_file = args['tissue_file']
    else:
        tissue_file = None

    if args['gzip'] == None:
        gzip = True
    elif args['gzip'] == 'True':
        gzip = True
    elif args['gzip'] == 'False':
        gzip = False
    else:
        print('Only True or False accepted for gzip')
        sys.exit(1)

    if args['neighborhood'] != None:
        try:
            neighborhood = int(args['neighborhood'])
        except ValueError:
            print('neighborhood has to be int')
            sys.exit(1)
    else:
        neighborhood = 4

    if args['feature_remove'] != None:
        if args['feature_remove'][-3:] != 'txt':
            feature_remove =  args['feature_remove'].split(',')
        else:
            with open(args['feature_remove'],'r') as f:
                feature_remove = [line.rstrip() for line in f]
    else:
        feature_remove = None

    if args['color'] != None:
        if args['color'] not in ['black','blue']:
            print('color should be black or blue')
            sys.exit(1)
        else:
            color =  args['color']
    else:
        color = 'blue'

    if args['line_width'] != None:
        try:
            linewidth = float(args['line_width'])
        except ValueError:
            print('line width has to be int or float')
            sys.exit(1)
    else:
        linewidth = None

    if args['feature_set'] != None:
        if args['feature_set'][-3:] != 'txt':
            feature_set =  args['feature_set'].split(',')
        else:
            with open(args['feature_set'],'r') as f:
                feature_set = [line.rstrip() for line in f]
    else:
        feature_set = None


    if verbose > 0:
        print("""
Parameters utilized:
target_file="""+args['target_file']+"""
feature_files="""+str(args['feature_files'].split(','))+"""
seed_files="""+str(seed_files)+"""
prefix="""+prefix+"""
seed_name="""+str(seed_name)+"""
grid="""+str(grid)+"""
target_name="""+str(target_name)+"""
k="""+str(k)+"""
bandwidth_mult="""+str(bandwidth_mult)+"""
bandwidth_adj="""+str(bandwidth_adj)+"""
direction="""+direction+"""
mode="""+mode+"""
num_top="""+str(num_top)+"""
low_threshold="""+str(low_threshold)+"""
high_threshold="""+str(high_threshold)+"""
collapse="""+str(collapse)+"""
normalize="""+normalize+"""
gene_locus="""+gene_locus+"""
verbose="""+str(verbose)+"""
max_iteration="""+str(max_iteration)+"""
thread_number="""+str(thread_number)+"""
figure_format="""+figure_format+"""
subset="""+str(subset)+"""
if_pval="""+str(if_pval)+"""
if_bootstrap="""+str(if_bootstrap)+"""
if_cluster="""+str(if_cluster)+"""
if_intermediate="""+str(if_intermediate)+"""
out_folder="""+out_folder+"""
gzip="""+str(gzip))

    #if grid_check == False:
    runREVEALER(target_file=args['target_file'], # gct file for target(continuous or binary)
                    feature_files=feature_files, # gct file for features(binary)
                    seed_files=seed_files, # file for seed, if not provided, feature file is used directly 
                    prefix=prefix, # prefix for result files 
                    seed_name=seed_name, # names for seed, should be a list of string indicating the name of seed
                    grid=grid, # number of grid, default is 34
                    target_name=target_name, # name/index of target in target file. can be int n for nth row, or string s for row with index s
                    k=k, # size of kernel for k standard deviation away
                    bandwidth_mult=bandwidth_mult, # multiplication for bandwidth
                    bandwidth_adj=bandwidth_adj, # adjustion value for bandwidth
                    direction=direction, # direction that feature should match with target
                    mode=mode, # indicate if multiple parameter set is passes. if True, then prefix, k, grid, bandwidth_mult, and bandwidth_adj has to be a list
                    num_top=num_top, # number of top matches shown in intermediate file
                    low_threshold=low_threshold, # lowest threshold that feature with less than this value occurence will be removed
                    high_threshold=high_threshold, # highest threshold that feature with less than this value occurence will be removed
                    collapse=collapse, # indicate if same features are collapsed together for intermediate files
                    normalize=normalize, # normalize method for target
                    gene_locus=gene_locus, # gene_locus file indicating gene name and location of that gene
                    verbose=verbose, # verbose level(if 0, no report)
                    max_iteration=max_iteration, # maximum of iteration for best CIC discovery, automatic detection by IC value if -1 
                    thread_number=thread_number, # number of core used for parallel computing.
                    figure_format=figure_format, # format for result figure
                    subset=subset, # if list of string passed, only columns in this list is picked for calculation
                    if_pval=if_pval, # if True, p-values are calculated for all result
                    if_bootstrap=if_bootstrap, # if True, variance by bootstrap is calculated for all result
                    if_cluster=if_cluster, # if True, features in intermediate files are clustered with NMF 
                    if_intermediate=if_intermediate, # if True, intermediate result with top CIC value features are reported
                    out_folder=out_folder, # folder to put output files inside
                    gene_set = gene_set, # set of gene to run REVEALER
                    gene_separator = separator,  #separator between gene and later part
                    gmt_file = gmt_file,
                    alpha = alpha,
                    tissue_file = tissue_file,
                    gzip = gzip,
                    neighborhood = neighborhood,
                    feature_remove = feature_remove,
                    color = color,
                    linewidth = linewidth,
                    feature_set = feature_set
                    )
    #else:
        # runCheckGrid(target_file=args['target_file'], # gct file for target(continuous or binary)
        #             feature_file=args['feature_file'], # gct file for features(binary)
        #             seed_file=seed_file, # file for seed, if not provided, feature file is used directly 
        #             prefix=prefix, # prefix for result files 
        #             seed_name=seed_name, # names for seed, should be a list of string indicating the name of seed
        #             grid=grid, # number of grid, default is 34
        #             target_name=target_name, # name/index of target in target file. can be int n for nth row, or string s for row with index s
        #             k=k, # size of kernel for k standard deviation away
        #             bandwidth_mult=bandwidth_mult, # multiplication for bandwidth
        #             bandwidth_adj=bandwidth_adj, # adjustion value for bandwidth
        #             direction=direction, # direction that feature should match with target
        #             mode=mode, # indicate if multiple parameter set is passes. if True, then prefix, k, grid, bandwidth_mult, and bandwidth_adj has to be a list
        #             num_top=num_top, # number of top matches shown in intermediate file
        #             low_threshold=low_threshold, # lowest threshold that feature with less than this value occurence will be removed
        #             high_threshold=high_threshold, # highest threshold that feature with less than this value occurence will be removed
        #             collapse=collapse, # indicate if same features are collapsed together for intermediate files
        #             normalize=normalize, # normalize method for target
        #             gene_locus=gene_locus, # gene_locus file indicating gene name and location of that gene
        #             verbose=verbose, # verbose level(if 0, no report)
        #             max_iteration=max_iteration, # maximum of iteration for best CIC discovery, automatic detection by IC value if -1 
        #             thread_number=thread_number, # number of core used for parallel computing.
        #             figure_format=figure_format, # format for result figure
        #             subset=subset, # if list of string passed, only columns in this list is picked for calculation
        #             if_pval=if_pval, # if True, p-values are calculated for all result
        #             if_bootstrap=if_bootstrap, # if True, variance by bootstrap is calculated for all result
        #             if_cluster=if_cluster, # if True, features in intermediate files are clustered with NMF 
        #             if_intermediate=if_intermediate, # if True, intermediate result with top CIC value features are reported
        #             out_folder=out_folder, # folder to put output files inside
        #             gene_set = gene_set, # set of gene to run REVEALER
        #             gene_separator = separator,  #separator between gene and later part
        #             gmt_file = gmt_file,
        #             alpha = alpha
        #             )




import argparse
from revealer.REVEALER_Cython import runREVEALER
from revealer.MutMaker import produce_mutation_file
from revealer.REVEALER_test import run_test
import sys

def main():
    parser = argparse.ArgumentParser(description='REVEALER: A tool for identifying significant mutually exclusive features.')
    subparsers = parser.add_subparsers(dest='subcommand')

    # Preprocess subcommand
    preprocess_parser = subparsers.add_parser('preprocess', help='Preprocessing MAF file for REVEALER input.')
    preprocess_required = preprocess_parser.add_argument_group('required arguments')
    preprocess_optional = preprocess_parser.add_argument_group('optional arguments')
    
    preprocess_required.add_argument("-i", "--input_file", dest="input_file", required=True,
                                     help="Name of input file, can be maf file or gct file.")
    preprocess_required.add_argument("-pi", "--protein_change_identifier", dest="protein_change_identifier", required=True,
                                     help="protein change identifier for dataset.")
    preprocess_required.add_argument("-cg", "--col_genename", dest="col_genename", required=True,
                                     help="column with gene name information.")
    preprocess_required.add_argument("-cc", "--col_class", dest="col_class", required=True,
                                     help="column with class information.")
    preprocess_required.add_argument("-csa", "--col_sample", dest="col_sample", required=True,
                                     help="column with sample name information.")
    
    preprocess_optional.add_argument("-m", "--mode", dest="mode", default='class',
                                     help="Mode, can be class, freq, weight, all, or comb. For more detail, check documentation. Default is class.")
    preprocess_optional.add_argument("-r", "--ratio", dest="ratio", type=float, default=1/3,
                                     help="Ratio of selected features by weight that is acceptable. Default is 1/3.")
    preprocess_optional.add_argument("-pf", "--phenotype_file", dest="phenotype_file", default=None,
                                     help="Name of phenotype file. Required if mode is one of freq, weight, all, or comb.")
    preprocess_optional.add_argument("-c", "--class_file", dest="class_file", default=None,
                                     help="File for classifiers, required if input file is gct file.")
    preprocess_optional.add_argument("-p", "--prefix", dest="prefix", default='Mut',
                                     help="prefix for output files. default is Mut.")
    preprocess_optional.add_argument("-o", "--out_folder", dest="out_folder", default='./',
                                     help="Path to directory to put output files. default is current directory.")
    preprocess_optional.add_argument("-cs", "--class_separator", dest="class_separator", default='_',
                                     help="Separator between gene name and class name. default is '_'")
    preprocess_optional.add_argument("-pn", "--phenotype_name", dest="phenotype_name", default=0,
                                     help="Name of phenotypr to be used in phenotype file. Default is first row.")
    preprocess_optional.add_argument("-fs", "--file_separator", dest="file_separator", default='\t',
                                     help="Separator of file, default is tab('\\t')")
    preprocess_optional.add_argument("-d", "--direction", dest="direction", default='pos',
                                     help="direction of phenotype the features should match. Default is positive.")
    preprocess_optional.add_argument("-ft", "--frequency_threshold", dest="frequency_threshold", type=int, default=5,
                                     help="Threshold for frequency. Default is 5.")
    preprocess_optional.add_argument("-wt", "--weight_threshold", dest="weight_threshold", type=float, default=0.7,
                                     help="Threshold for weight. Default is 0.7.")
    preprocess_optional.add_argument("-gl", "--gene_list", dest="gene_list", default=None,
                                     help="List of Gene to run. Can be file or ',' separated names. Default is all gene.")
    preprocess_optional.add_argument("-nm", "--name_match", dest="name_match", type=str, choices=['True', 'False'], default='True',
                                     help="Indicate if sample name in phenotype file and input file are matching. Default is True.")
    preprocess_optional.add_argument("-mf", "--make_figure", dest="make_figure", type=str, choices=['True', 'False'], default='False',
                                     help="Indicate if figures are created. Recommended only when you run few genes. Default is False.")
    preprocess_optional.add_argument("-ff", "--figure_format", dest="figure_format", default='pdf',
                                     help="Format for figure. Default is pdf.")
    preprocess_optional.add_argument("-sl", "--sample_list", dest="sample_list", default=None,
                                     help="List of sample to be used.")
    preprocess_optional.add_argument("-tr", "--total_ratio", dest="total_ratio", type=float, default=0.4,
                                     help="Ratio of gene occrence compared to total sample number. default is 0.4")
    preprocess_optional.add_argument("-ig", "--if_gmt", dest="if_gmt", type=str, choices=['True', 'False'], default='True',
                                     help="Indicate if gmt file should be generated. Default is True.")
    preprocess_optional.add_argument("-v", "--verbose", dest="verbose", type=int, default=1,
                                     help="Indicate level of verbose. default is 1.")
    preprocess_optional.add_argument("-k", "--k_size", dest="k", type=int, default=30,
                                     help="Size of kernel indicating k variance far from middle. Has to be int. Default is 5")
    preprocess_optional.add_argument("-bm", "--bandwidth_multiplication", dest="bandwidth_mult", type=float, default=0.65,
                                     help="Value of bandwidth multiplier. Has to be float or int. Default is 0.65.")
    preprocess_optional.add_argument("-ba", "--bandwidth_adjustment", dest="bandwidth_adj", type=float, default=-0.95,
                                     help="Value of bandwidth adjustment. Has to be float or int. Default is -0.95")
    preprocess_optional.add_argument("-gz", "--gzip", dest="gzip", type=str, choices=['True', 'False'], default='True',
                                     help="If output file should be gzipped.")
    preprocess_optional.add_argument("-cb", "--combine", dest="combine", type=str, choices=['True', 'False'], default='True',
                                     help="Combine alleles by gene when making figures")
    preprocess_optional.add_argument("-nh", "--neighborhood", dest="neighborhood", type=int, default=4,
                              help="number of neighborhood.")

    # Run subcommand
    run_parser = subparsers.add_parser('run', help='Run REVEALER analysis.')
    run_required = run_parser.add_argument_group('required arguments')
    run_optional = run_parser.add_argument_group('optional arguments')

    run_required.add_argument("-tf", "--target_file", dest="target_file", required=True,
                              help="Name of target file, has to be gct file.")
    run_required.add_argument("-f", "--feature_files", dest="feature_files", required=True,
                              help="Name of feature file, has to be gct file.")

    run_optional.add_argument("-sf", "--seed_files", dest="seed_files", default=None,
                              help="Name of seed file.")
    run_optional.add_argument("-p", "--prefix", dest="prefix", default='REVEALER',
                              help="Prefix for results files. Default is REVEALER.")
    run_optional.add_argument("-sn", "--seed_name", dest="seed_name", default=None,
                              help="Seed names to take from seed file or feature file. If multiple exist, separate by ','. Passing text file(ending with 'txt') with column of seed names also works")
    run_optional.add_argument("-g", "--grid", dest="grid", type=int, default=34,
                              help="Grid size to put kernel on. has to be int. Default is 34.")
    run_optional.add_argument("-t", "--target_name", dest="target_name", default=0,
                              help="Name of target in target file. Can be row name or row index(0 based). Default is first row.")
    run_optional.add_argument("-k", "--k_size", dest="k", type=int, default=5,
                              help="Size of kernel indicating k variance far from middle. Has to be int. Default is 5")
    run_optional.add_argument("-bm", "--bandwidth_multiplication", dest="bandwidth_mult", type=float, default=0.65,
                              help="Value of bandwidth multiplier. Has to be float or int. Default is 0.65.")
    run_optional.add_argument("-ba", "--bandwidth_adjustment", dest="bandwidth_adj", type=float, default=-0.95,
                              help="Value of bandwidth adjustment. Has to be float or int. Default is -0.95")
    run_optional.add_argument("-d", "--direction", dest="direction", default='pos',
                              help="Direction of phenotype the features should match. Default is positive.")
    run_optional.add_argument("-m", "--mode", dest="mode", default='single',
                              help="Mode to run REVEALER. Can be single or multiple. If set to multiple, format as sample input is required. Default is single.")
    run_optional.add_argument("-ps", "--parameter_set", dest="parameter_set", default=None,
                              help="set of parameter. Required if mode is multiple. Check available format in documentation.")
    run_optional.add_argument("-nt", "--num_top", dest="num_top", type=int, default=30,
                              help="Number of top features picked for intermediate report. Has to be int. Default is 30.")
    run_optional.add_argument("-lt", "--low_threshold", dest="low_threshold", type=float, default=3,
                              help="Lower threshold to remove feature. If int passed, threshold of absolute value is made. If float passed, threshold made by ratio. Default is 3.")
    run_optional.add_argument("-ht", "--high_threshold", dest="high_threshold", type=float, default=100,
                              help="Higher threshold to remove feature. If int passed, threshold of absolute value is made. If float passed, threshold made by ratio. Default is 0.2.")
    run_optional.add_argument("-ic", "--if_collapse", dest="collapse", type=str, choices=['True', 'False'], default='False',
                              help="Indicate if features are collapsed for intermediate report. Default is False.")
    run_optional.add_argument("-ff", "--figure_format", dest="figure_format", default='pdf',
                              help="Format for figure. Can be any figure format available in matplotlib. Default is pdf.")
    run_optional.add_argument("-tn", "--thread_number", dest="thread_number", type=int, default=1,
                              help="Number of thread used to run the program. Default is 1, -1 indicate all thread used.")
    run_optional.add_argument("-n", "--normalize", dest="normalize", default='zerobase',
                              help="Way to normalize the target. Can be standard or zerobase. Default is zerobase.")
    run_optional.add_argument("-gl", "--gene_locus", dest="gene_locus", default='None',
                              help="File name to indicate gene locus. Check documentation for format. Default is None.")
    run_optional.add_argument("-v", "--verbose", dest="verbose", type=int, default=1,
                              help="Verbose level for report. Setting 0 to get no report. Default is 1.")
    run_optional.add_argument("-mi", "--max_iteration", dest="max_iteration", type=int, default=-1,
                              help="Number of iteration. Has to be int. -1 indicates stop by automatic detection. Default is -1")
    run_optional.add_argument("-s", "--subset", dest="subset", default='no',
                              help="Subset of samples to be chosen. Has to be file with one column with subset of sample names in target file. Default is use all columns.")
    run_optional.add_argument("-ip", "--if_pvalue", dest="if_pval", type=str, choices=['True', 'False'], default='True',
                              help="Indicate if pvalues are calculated and plotted on figure. Default is True.")
    run_optional.add_argument("-ib", "--if_bootstrap", dest="if_bootstrap", type=str, choices=['True', 'False'], default='True',
                              help="Indicate if variance are calculated by bootstrap and plotted on figure. Default is True.")
    run_optional.add_argument("-icl", "--if_cluster", dest="if_cluster", type=str, choices=['True', 'False'], default='False',
                              help="Indicate if features are clustered using NMF for intermediate report. Default is False.")
    run_optional.add_argument("-ii", "--if_intermediate", dest="if_intermediate", type=str, choices=['True', 'False'], default='True',
                              help="Indicate if intermediate reports are generated. Default is True.")
    run_optional.add_argument("-o", "--out_folder", dest="out_folder", default='./',
                              help="Path to directory to put output files. default is current directory.")
    run_optional.add_argument("-sep", "--separator", dest="separator", default='_',
                              help="Separator between gene name and later part.")
    run_optional.add_argument("-gs", "--gene_set", dest="gene_set", default=None,
                              help="Gene Set that is extracted.")
    run_optional.add_argument("-gmt", "--gmt_file", dest="gmt_file", default=None,
                              help="gmt file for allele information.")
    run_optional.add_argument("-a", "--alpha", dest="alpha", type=float, default=1,
                              help="Power to raise value. Default is 1.")
    run_optional.add_argument("-gc", "--grid_check", dest="grid_check", type=str, choices=['True', 'False'], default='False',
                              help="Indicate if run Grid Check mode. Default is False.")
    run_optional.add_argument("-tif", "--tissue_file", dest="tissue_file", default=None,
                              help="File of tissue information.")
    run_optional.add_argument("-gz", "--gzip", dest="gzip", type=str, choices=['True', 'False'], default='False',
                              help="If output file should be gzipped.")
    run_optional.add_argument("-nh", "--neighborhood", dest="neighborhood", type=int, default=4,
                              help="number of neighborhood.")
    run_optional.add_argument("-fr", "--feature_remove", dest="feature_remove", default=None,
                              help="features to remove manually.")
    run_optional.add_argument("-c", "--color", dest="color", default='blue',
                              help="tune of color palette.")
    run_optional.add_argument("-lw", "--line_width", dest="line_width", type=float, default=None,
                              help="width of line.")
    run_optional.add_argument("-fs", "--feature_set", dest="feature_set", default=None,
                              help="set of feature to use.")

    test_parser = subparsers.add_parser('test', help='Testing the work of REVEALER after download.')

    args = parser.parse_args()

    if args.subcommand == 'preprocess':
        preprocess(args)
    elif args.subcommand == 'run':
        run(args)
    elif args.subcommand == 'test':
        run_test(args)
    else:
        parser.print_help()

def preprocess(args):
    if args.input_file[-3:] != 'gct' and args.input_file[-3:] != 'maf' and args.input_file[-3:] != 'csv':
        print('only maf file and gct file are accepted as input, please check your input file.')
        sys.exit(1)
    else:
        input_file = args.input_file

    if args.mode not in ['mutall', 'allele', 'class', 'freq', 'weight', 'weight_filter', 'comb']:
        print('Mode not available! Please check instruction for available mode.')
        sys.exit(1)

    if (args.mode in ['freq', 'weight', 'weight_filter', 'comb']) and args.phenotype_file is None:
        print('Please indicate phenotype file if use mode ' + args.mode)
        sys.exit(1)

    if args.input_file[-3:] == 'gct' and args.class_file is None:
        print('Please indicate class names by passing class file.')
        sys.exit(1)

    if args.input_file[-3:] == 'gct':
        with open(args.class_file, 'r') as f:
            class_list = [line.rstrip() for line in f]
    else:
        class_list = ['Nonsense_Mutation', 'In_Frame_Del', 'Silent', 'Frame_Shift_Ins', 'Missense_Mutation', 'Splice_Site',
                      'Frame_Shift_Del', 'De_novo_Start_OutOfFrame', 'Nonstop_Mutation', 'In_Frame_Ins', 'Start_Codon_SNP',
                      'Start_Codon_Del', 'Stop_Codon_Ins', 'Start_Codon_Ins', 'Stop_Codon_Del', 'Intron', 'IGR', "5'Flank",
                      "3'UTR", "5'UTR", 'Mut_All']

    if args.gene_list is not None:
        if args.gene_list[-3:] != 'txt':
            gene_list = args.gene_list.split(',')
        else:
            with open(args.gene_list, 'r') as f:
                gene_list = [line.rstrip() for line in f]
    else:
        gene_list = None

    if args.sample_list is not None:
        with open(args.sample_list, 'r') as f:
            sample_list = [line.rstrip() for line in f]
    else:
        sample_list = None

    if args.verbose > 0:
        print(f"""
    Parameters utilized:
    input_file={input_file}
    protein_change_identifier={args.protein_change_identifier}
    mode={args.mode}
    class_list={str(class_list)}
    prefix={args.prefix}
    ratio={str(args.ratio)}
    out_folder={str(args.out_folder)}
    class_separator={str(args.class_separator)}
    phenotype_name={str(args.phenotype_name)}
    file_separator={str(args.file_separator)}
    direction={args.direction}
    frequency_threshold={str(args.frequency_threshold)}
    weight_threshold={str(args.weight_threshold)}
    total_ratio={str(args.total_ratio)}
    gene_list={str(gene_list)}
    sample_list={str(sample_list)}
    name_match={str(args.name_match)}
    make_figure={str(args.make_figure)}
    figure_format={str(args.figure_format)}
    if_gmt={str(args.if_gmt)}
    verbose={str(args.verbose)}
    k={str(args.k)}
    bandwidth_mult={str(args.bandwidth_mult)}
    bandwidth_adj={str(args.bandwidth_adj)}
    gzip={str(args.gzip)}
    combine={str(args.combine)}
    neighborhood={str(args.neighborhood)}
    """)

    produce_mutation_file(
        maf_input_file=input_file if (input_file[-3:] == 'maf' or input_file[-3:] == 'csv') else None,
        gct_input_file=input_file if (input_file[-3:] == 'gct' ) else None,
        gct_output_file_prefix=args.prefix,
        phenotype_file=args.phenotype_file,
        class_list=class_list,
        class_seperator=args.class_separator,
        phenotype_name=args.phenotype_name,
        file_separator=args.file_separator,
        protein_change_identifier=args.protein_change_identifier,
        mode=args.mode,
        direction=args.direction,
        frequency_threshold=args.frequency_threshold,
        weight_threshold=args.weight_threshold,
        gene_list=gene_list,
        name_match=args.name_match == 'True',
        make_figure=args.make_figure == 'True',
        figure_format=args.figure_format,
        out_folder=args.out_folder,
        ratio=args.ratio,
        sample_list=sample_list,
        total_ratio=args.total_ratio,
        if_gmt=args.if_gmt == 'True',
        k=args.k,
        gzip=args.gzip == 'True',
        combine=args.combine == 'True',
        col_genename=args.col_genename,
        col_class=args.col_class,
        col_sample=args.col_sample,
        neighborhood=args.neighborhood,
    )

def run(args):
    def process_file_input(file_input):
        if file_input.endswith('.gct'):
            return [file_input]
        elif file_input.endswith('.txt'):
            with open(file_input, 'r') as f:
                return [line.strip() for line in f if line.strip() and line.strip().endswith('.gct')]
        else:
            return [f.strip() for f in file_input.split(',') if f.strip().endswith('.gct')]

    feature_files = process_file_input(args.feature_files)
    if not feature_files:
        print('No valid .gct files found in the feature files input.')
        sys.exit(1)

    if args.gmt_file is not None:
        if args.gmt_file[-3:] != 'gmt':
            print('gmt file has to be in gmt format.')
            sys.exit(1)
        else:
            gmt_file = args.gmt_file
    else:
        gmt_file = None

    if args.seed_files is not None:
        seed_files = process_file_input(args.seed_files)
        if not seed_files:
            print('No valid .gct files found in the seed files input.')
            sys.exit(1)
    else:
        seed_files = feature_files

    if args.seed_name is not None:
        if args.seed_name[-3:] != 'txt':
            seed_name = args.seed_name.split(',')
        else:
            with open(args.seed_name, 'r') as f:
                seed_name = [line.rstrip() for line in f]
    else:
        seed_name = None

    if args.direction == 'positive':
        direction = 'pos'
    elif args.direction == 'negative':
        direction = 'neg'
    else:
        print('Direction has to be either positive or negative')
        sys.exit(1)

    if args.mode == 'multiple':
        if args.parameter_set is None:
            print('Please specify parameter set if use mode multiple.')
            sys.exit(1)
        else:
            if args.parameter_set[-3:] == 'txt':
                params = pd.read_csv(args.parameter_set, sep='\t')
                prefix = params['prefix'].tolist()
                grid = params['grid'].tolist()
                k = params['k'].tolist()
                bandwidth_mult = params['bandwidth_mult'].tolist()
                bandwidth_adj = params['bandwidth_adj'].tolist()
            else:
                paramall = args.parameter_set.split(';')
                paramlist = []
                for i in paramall:
                    paramlist.append(i.split(','))
                params = pd.DataFrame(paramlist, columns=['prefix', 'grid', 'k', 'bandwidth_mult', 'bandwidth_adj'])
                prefix = params['prefix'].tolist()
                grid = params['grid'].tolist()
                k = params['k'].tolist()
                bandwidth_mult = params['bandwidth_mult'].tolist()
                bandwidth_adj = params['bandwidth_adj'].tolist()
    else:
        prefix = args.prefix
        grid = args.grid
        k = args.k
        bandwidth_mult = args.bandwidth_mult
        bandwidth_adj = args.bandwidth_adj

    if args.subset != 'no':
        with open(args.subset, 'r') as f:
            subset = [line.rstrip() for line in f]
    else:
        subset = args.subset

    if args.gene_set is not None:
        if args.gene_set[-3:] != 'txt':
            gene_set = args.gene_set.split(',')
        else:
            with open(args.gene_set, 'r') as f:
                gene_set = [line.rstrip() for line in f]
    else:
        gene_set = None

    if args.feature_remove is not None:
        if args.feature_remove[-3:] != 'txt':
            feature_remove = args.feature_remove.split(',')
        else:
            with open(args.feature_remove, 'r') as f:
                feature_remove = [line.rstrip() for line in f]
    else:
        feature_remove = None

    if args.feature_set is not None:
        if args.feature_set[-3:] != 'txt':
            feature_set = args.feature_set.split(',')
        else:
            with open(args.feature_set, 'r') as f:
                feature_set = [line.rstrip() for line in f]
    else:
        feature_set = None

    if args.verbose > 0:
        print(f"""
    Parameters utilized:
    target_file={args.target_file}
    feature_files={str(feature_files)}
    seed_files={str(seed_files)}
    prefix={prefix}
    seed_name={str(seed_name)}
    grid={str(grid)}
    target_name={str(args.target_name)}
    k={str(k)}
    bandwidth_mult={str(bandwidth_mult)}
    bandwidth_adj={str(bandwidth_adj)}
    direction={direction}
    mode={args.mode}
    num_top={str(args.num_top)}
    low_threshold={str(args.low_threshold)}
    high_threshold={str(args.high_threshold)}
    collapse={str(args.collapse)}
    normalize={args.normalize}
    gene_locus={args.gene_locus}
    verbose={str(args.verbose)}
    max_iteration={str(args.max_iteration)}
    thread_number={str(args.thread_number)}
    figure_format={args.figure_format}
    subset={str(subset)}
    if_pval={str(args.if_pval)}
    if_bootstrap={str(args.if_bootstrap)}
    if_cluster={str(args.if_cluster)}
    if_intermediate={str(args.if_intermediate)}
    out_folder={args.out_folder}
    gzip={str(args.gzip)}
    """)

    runREVEALER(target_file=args.target_file,
                feature_files=feature_files,
                seed_files=seed_files,
                prefix=prefix,
                seed_name=seed_name,
                grid=grid,
                target_name=args.target_name,
                k=k,
                bandwidth_mult=bandwidth_mult,
                bandwidth_adj=bandwidth_adj,
                direction=direction,
                mode=args.mode,
                num_top=args.num_top,
                low_threshold=args.low_threshold,
                high_threshold=args.high_threshold,
                collapse=args.collapse == 'True',
                normalize=args.normalize,
                gene_locus=args.gene_locus,
                verbose=args.verbose,
                max_iteration=args.max_iteration,
                thread_number=args.thread_number,
                figure_format=args.figure_format,
                subset=subset,
                if_pval=args.if_pval == 'True',
                if_bootstrap=args.if_bootstrap == 'True',
                if_cluster=args.if_cluster == 'True',
                if_intermediate=args.if_intermediate == 'True',
                out_folder=args.out_folder,
                gene_set=gene_set,
                gene_separator=args.separator,
                gmt_file=gmt_file,
                alpha=args.alpha,
                tissue_file=args.tissue_file,
                gzip=args.gzip == 'True',
                neighborhood=args.neighborhood,
                feature_remove=feature_remove,
                color=args.color,
                linewidth=args.line_width,
                feature_set=feature_set)


if __name__ == '__main__':
    main()



from numpy import divide, dot, finfo, log, matrix, multiply, ndarray
from numpy.random import rand, seed
from sklearn.decomposition import NMF

RANDOM_SEED = 20121020


def nmf(a,
        k,
        algorithm='als',
        init=None,
        solver='cd',
        beta_loss='frobenius',
        tol=0.0001,
        max_iter=200,
        random_seed=RANDOM_SEED,
        alpha=0.0,
        l1_ratio=0.0,
        verbose=0,
        shuffle=False):
    """
    Non-negative matrix factorize a using k: A ~ W * H.
    Arguments:
        a (array): (n_samples, n_features); A matrix
        k (int): number of hidden variables
        algorithm (str): 'als' | 'ls'
        init:
        solver:
        beta_loss:
        tol:
        max_iter:
        random_seed (int | array):
        alpha:
        l1_ratio:
        verbose:
        shuffle:
    Returns:
        array: W matrix
        array: H matrix
        float: reconstruction error
    """

    if algorithm == 'als':

        model = NMF(
            n_components=k,
            init=init,
            solver=solver,
            beta_loss=beta_loss,
            tol=tol,
            max_iter=max_iter,
            random_state=random_seed,
            alpha=alpha,
            l1_ratio=l1_ratio,
            verbose=verbose,
            shuffle=shuffle)

        w = model.fit_transform(a)
        h = model.components_
        e = model.reconstruction_err_

    elif algorithm == 'ls':

        w, h, e = _nmf_div(
            a, k, n_iterations=max_iter, random_seed=random_seed)

    else:
        raise ValueError('Unknown algorithm: {}.'.format(algorithm))

    return w, h, e


def _nmf_div(a, k, n_iterations=1000, random_seed=RANDOM_SEED):
    """
    Non-negative matrix factorize matrix using k with divergence algorithm.
    Arguments:
        a (array): (n_samples, n_features); A matrix
        k (int): number of hidden variables
        n_max_iterations (int):
        random_seed (int | array):
    Returns:
        array: W matrix
        array: H matrix
        float: reconstruction error
    """

    # TODO: optimize

    eps = finfo(float).eps

    n = a.shape[0]
    m = a.shape[1]
    a = matrix(a)

    seed(random_seed)
    w = rand(n, k)
    h = rand(k, m)

    for t in range(n_iterations):

        a_p = dot(w, h)

        w_t = matrix.transpose(w)

        h = multiply(h, dot(w_t, divide(a, a_p))) + eps

        for i in range(k):
            w_sum = 0
            for j in range(n):
                w_sum += w[j, i]
            for j in range(m):
                h[i, j] = h[i, j] / w_sum

        a_p = dot(w, h)

        h_t = matrix.transpose(h)

        w = multiply(w, dot(divide(a, a_p + eps), h_t)) + eps

        w = divide(w, ndarray.sum(h, axis=1, keepdims=False))

    e = (multiply(a, log(divide(a + eps, a_p + eps))) - a + a_p).sum() / (
        m * n)

    return w, h, e



from matplotlib.pyplot import gca, sca, suptitle
from seaborn import despine, set_style

from revealer.ccal.style import FONT_LARGER, FONT_LARGEST, FONT_SMALLER, FONT_STANDARD
from revealer.ccal.dict_ import merge_dicts_with_function


def decorate(ax=None,
             style='ticks',
             despine_kwargs={},
             title=None,
             title_kwargs={},
             yaxis_position='left',
             xlabel=None,
             ylabel=None,
             xlabel_kwargs={},
             ylabel_kwargs={},
             xticks=None,
             yticks=None,
             max_n_xticks=None,
             max_n_yticks=None,
             max_xtick_size=None,
             max_ytick_size=None,
             xticklabels_kwargs={},
             yticklabels_kwargs={},
             legend_loc='best'):
    """
    Decorate an ax.
    """

    if ax:
        sca(ax)
    else:
        ax = gca()

    if legend_loc:
        ax.legend(
            loc=legend_loc,
            prop={
                'size': FONT_STANDARD['fontsize'],
                'weight': FONT_STANDARD['weight'],
            },
            labels=['values'])

    # Set plot style
    set_style(style)
    despine(**despine_kwargs)

    # Title
    if title:
        title_kwargs = merge_dicts_with_function(FONT_LARGEST, title_kwargs,
                                                 lambda a, b: b)
        suptitle(title, **title_kwargs)

    # Set y axis position
    if yaxis_position == 'right':
        ax.yaxis.tick_right()

    # Style x label
    if xlabel is None:
        xlabel = ax.get_xlabel()
    xlabel_kwargs = merge_dicts_with_function(FONT_LARGER, xlabel_kwargs,
                                              lambda a, b: b)
    ax.set_xlabel(xlabel, **xlabel_kwargs)

    # Style y label
    if ylabel is None:
        ylabel = ax.get_ylabel()
    ylabel_kwargs = merge_dicts_with_function(FONT_LARGER, ylabel_kwargs,
                                              lambda a, b: b)
    ax.set_ylabel(ylabel, **ylabel_kwargs)

    # Style x ticks
    if xticks is not None:
        ax.set_xticks(xticks)

    if max_n_xticks and max_n_xticks < len(ax.get_xticks()):
        ax.set_xticks([])

    # Style x tick labels
    xticklabels = [t.get_text() for t in ax.get_xticklabels()]
    if len(xticklabels):

        if xticklabels[0]:
            # Limit tick label size
            if max_xtick_size:
                xticklabels = [t[:max_xtick_size] for t in xticklabels]
        else:
            xticklabels = ax.get_xticks()

        # Set tick label rotation
        if 'rotation' not in xticklabels_kwargs:
            xticklabels_kwargs['rotation'] = 90

        xticklabels_kwargs = merge_dicts_with_function(
            FONT_SMALLER, xticklabels_kwargs, lambda a, b: b)

        ax.set_xticklabels(xticklabels, **xticklabels_kwargs)

    # Style y ticks
    if yticks is not None:
        ax.set_yticks(yticks)

    if max_n_yticks and max_n_yticks < len(ax.get_yticks()):
        ax.set_yticks([])

    # Style y tick labels
    yticklabels = [t.get_text() for t in ax.get_yticklabels()]
    if len(yticklabels):

        if yticklabels[0]:
            # Limit tick label size
            if max_ytick_size:
                yticklabels = [t[:max_ytick_size] for t in yticklabels]
        else:
            yticklabels = ax.get_yticks()

        # Set tick label rotation
        if 'rotation' not in yticklabels_kwargs:
            yticklabels_kwargs['rotation'] = 0

        yticklabels_kwargs = merge_dicts_with_function(
            FONT_SMALLER, yticklabels_kwargs, lambda a, b: b)

        ax.set_yticklabels(yticklabels, **yticklabels_kwargs)



from matplotlib.colors import ListedColormap
from numpy import number

from revealer.ccal.style import CMAP_CATEGORICAL_TAB20


def assign_colors(objects, cmap=CMAP_CATEGORICAL_TAB20, colors=()):
    """
    Assign colors to objects.
    Arguments:
        objects (iterable):
        cmap (matplotlib.cm):
        colors (iterable):
    Returns:
        dict: {state: color}
    """

    o_to_color = {}

    for i, o in enumerate(sorted(set(objects))):

        if len(colors):

            # Make colormap from colors
            cmap = ListedColormap(colors, N=len(colors))

        if isinstance(o, number):

            object_range = max(objects) - min(objects)

            if object_range:
                o_01 = (o - min(objects)) / object_range
                i = int(o_01 * cmap.N)

            else:
                i = 0

        c = cmap(i)

        o_to_color[o] = c

    return o_to_color



from matplotlib.pyplot import figure, show
import matplotlib.pyplot as plt

from revealer.ccal.decorate import decorate
from revealer.ccal.save_plot import save_plot
from revealer.ccal.style import FIGURE_SIZE


def plot_points(*args,
                ax=None,
                figure_size=FIGURE_SIZE,
                title='',
                xlabel='',
                ylabel='',
                file_path=None,
                **kwargs):
    """
    Plot points.
    """

    if ax:
        save_and_show = False
    else:
        ax = figure(figsize=figure_size).gca()
        save_and_show = True

    for k, v in [
        ('linestyle', ''),
        ('marker', '.'),
        ('color', '#20D9BA'),
    ]:
        if k not in kwargs:
            kwargs[k] = v

    ax.plot(*args, **kwargs)

    decorate(title=title, xlabel=xlabel, ylabel=ylabel)

    if save_and_show:

        if file_path:
            save_plot(file_path)

        plt.close()



from multiprocessing.pool import Pool

from numpy.random import seed


def multiprocess(function, args, n_jobs, random_seed=None):
    """
    Call function with args across n_jobs processes (n_jobs doesn't have to be
        the length of list_of_args).
    Arguments:
        function (callable):
        args (iterable): an iterable of [(1,2), (3, 4)] results in
            [function(1,2), function(3,4)]
        n_jobs (int): 0 < n_jobs
        random_seed (int | array):
    Returns:
        list:
    """

    if random_seed is not None:
        # Each process initializes with the current jobs' randomness (random
        # state & random state index). Any changes to these processes'
        # randomnesses won't update the current process' randomness.
        seed(random_seed)

    with Pool(n_jobs) as p:
        return p.starmap(function, args)



import sys
import pandas as pd

def main():
    if len(sys.argv) != 4:
        print("Usage: python DepMapToCCLE.py <NameConvert.csv> <CCLE_class.gct> <output_filename.gct>")
        sys.exit(1)

    # Read the file paths from command line arguments
    name_convert_file, feature_file, outfile = sys.argv[1], sys.argv[2], sys.argv[3]

    # Read feature file
    feature = pd.read_csv(feature_file,skiprows=[0,1],sep='\t',index_col=0)
    feature = feature.drop(columns=feature.columns[0])

    # Read the dat1a
    converter = pd.read_csv(name_convert_file)

    # Create a mapping dictionary from 'ModelID' to 'CCLEName'
    model_to_ccle = dict(zip(converter['ModelID'], converter['CCLEName']))

    # Rename the columns in ccle_class_df using the mapping
    feature.rename(columns=model_to_ccle, inplace=True)

    # Save the result
    with open(outfile, mode = "w") as output_file:
        output_file.writelines("#1.2\n{}\t{}\n".format(feature.shape[0], feature.shape[1] - 1))
        feature.to_csv(output_file, sep= '\t')

if __name__ == "__main__":
    main()

def run(name_convert_file, feature_file, outfile):
    # Read feature file
    feature = pd.read_csv(feature_file,skiprows=[0,1],sep='\t',index_col=0)
    feature = feature.drop(columns=feature.columns[0])

    # Read the dat1a
    converter = pd.read_csv(name_convert_file)

    # Create a mapping dictionary from 'ModelID' to 'CCLEName'
    model_to_ccle = dict(zip(converter['ModelID'], converter['CCLEName']))

    # Rename the columns in ccle_class_df using the mapping
    feature.rename(columns=model_to_ccle, inplace=True)

    # Save the result
    with open(outfile, mode = "w") as output_file:
        output_file.writelines("#1.2\n{}\t{}\n".format(feature.shape[0], feature.shape[1] - 1))
        feature.to_csv(output_file, sep= '\t')



from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext

import numpy

setup(cmdclass = {'build_ext': build_ext},
      ext_modules=cythonize("./REVEALER_Cython.pyx"),
      include_dirs=[numpy.get_include()])


#-f -a --compile-args=-DCYTHON_TRACE_NOGIL=1 
cimport cython
cython: infer_types=True
cython: profile=True
import warnings
warnings.filterwarnings("ignore")

from revealer.ccal import * 
import numpy as np
cimport numpy as np
import math
cdef np.double_t EPS = np.finfo(float).eps
import sys

import pandas as pd

from libc.float cimport DBL_EPSILON
from libc.limits cimport INT_MAX
from libc.math cimport exp, log, sqrt, M_PI, fabs, isnan, fmin, fmax, round, isnan
from libc.stdlib cimport rand, malloc, free, srand
from libc.stdio cimport printf
from cython.parallel import prange

cdef extern from "limits.h":
    int RAND_MAX
cdef int RANDOM_SEED = 20121020

cdef struct density:
    np.double_t d
    np.uint32_t* cnt
    
cdef np.double_t DELMAX=1000

import matplotlib.pyplot as plt
import plotly.express as px
from scipy.stats import norm
import seaborn as sns
from tqdm.notebook import tqdm
import time
import matplotlib.colors as clr
import multiprocessing as mp
import warnings
import pickle
import random
import os

plt.switch_backend('agg')

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef inline density VR_den_bin(
    int n,              #Size of example
    int nb,             #nb to calculate bcv
    np.double_t* x,     #Pointer to list x
    np.double_t x_min,  #Minumum value of list x
    np.double_t x_max   #Maximum value of list x
    ) nogil:

    '''
    Function to calculate bin dencity
    '''

    cdef int i, j, ii, jj, iij
    cdef density ret
    ret.cnt = <np.uint32_t*> malloc(nb * sizeof(np.uint32_t))
    for i in range(nb):
        ret.cnt[i] = 0
    
    cdef np.double_t rang = (x_max - x_min) * 1.01
    ret.d = rang / <np.double_t>nb

    cdef np.double_t x_i, x_j
    for i in range(1,n):
        ii = (int)(x[i] / ret.d)
        for j in range(i):
            jj = (int)(x[j] / ret.d)
            iij = (ii - jj)
            iij = iij if iij >= 0 else -iij
            if(ret.cnt[iij] == INT_MAX):
                printf("maximum count exceeded in pairwise distance binning\n")
            ret.cnt[iij] += 1
    return ret

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef inline np.double_t VR_bcv_bin(
    int n,          #N value
    int nb,         #nb value to calculate bcv
    density den,    #density
    np.double_t h   #h value
    ) nogil:

    '''
    Funnction to calculate bcv
    '''

    cdef np.double_t total_sum = 0.0
    cdef np.double_t delta, term
    cdef np.double_t hh = h / 4.0
    cdef int i
    for i in range(nb):
        delta = (i * den.d / hh) 
        delta *= delta
        if delta >= DELMAX:
            break
        term = exp(-1.0 * delta / 4.0) * (delta * delta - 12.0 * delta + 12.0)
        total_sum += term * den.cnt[i]
    return 1.0 / (2.0 * n * hh * sqrt(M_PI)) + total_sum / (64.0 * n * n * hh * sqrt(M_PI))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef inline np.double_t r_var(
    np.double_t* x,     #Pointer to list
    int n               #N value
    ) nogil:
    
    '''
    Function to calculate variance
    '''

    cdef np.double_t mean = 0.0
    cdef int i
    for i in range(n):
        mean += x[i]
    mean = mean / n
    
    cdef np.double_t out = 0.0
    
    for i in range(n):
        out += (x[i] - mean) ** 2
    return out / (n - 1.0)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef inline np.double_t sign(
    np.double_t x   #Value to be get sign
    ) nogil:

    '''
    Function to get sign of a double value
    '''

    if x < 0:
        return -1.0
    elif x == 0.0:
        return 0.0
    else:
        return 1.0
    
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline int imin(
    int x,  #X value 
    int y   #Y value
    ) nogil:

    '''
    Function to compare two value and return smaller one
    '''

    if x < y:
        return x
    else: 
        return y

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef inline int imax(
    int x,  #X value 
    int y   #Y value
    ) nogil:

    '''
    Function to compare two value and return larger one
    '''

    if x > y:
        return x
    else: 
        return y
    
@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline int rint(
    np.double_t x   #Value to be converted to int
    ) nogil:

    '''
    Functionn to make double to int
    '''

    return <int> round(x)

@cython.boundscheck(False)
@cython.initializedcheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.double_t findmin(
    np.double_t[:] arr,     #List to find minimum value
    int size                #Size of list
    ) nogil:

    '''
    Function to find minimum value in a list
    '''

    cdef np.double_t minVal = arr[0]
    cdef int i = 0
    for i in range(size):
        if arr[i] < minVal:
            minVal = arr[i]
    return minVal

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cdef np.double_t findmax(
    np.double_t[:] arr,     #List to find minimum value 
    int size                #Size of list
    ) nogil:

    '''
    Function to find maximum value in a list
    '''

    cdef np.double_t minVal = arr[0]
    cdef int i = 0
    for i in range(size):
        if arr[i] > minVal:
            minVal = arr[i]
    return minVal

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cpdef np.double_t calc_bandwidth(
    np.double_t[:] x_view,  #List of double
    int n,                  #Size of list
    int nb=1000             #nb value to calculate bcv
    ) nogil:

    '''
    Function to calculate bandwidth
    '''

    cdef int i = 0
    cdef np.double_t x_min = findmin(x_view,n)
    cdef np.double_t x_max = findmax(x_view,n)
    cdef np.double_t* x_ptr = &x_view[0]
    cdef np.double_t ret = 0.0
    
    ret = py_bcv(x_ptr, n, x_min, x_max, nb)
    return ret

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
@cython.cdivision(True)
cdef inline np.double_t py_bcv(
    np.double_t* x,     #Pointer to list x
    int n,              #Size of sample
    np.double_t x_min,  #Minimum value in list x
    np.double_t x_max,  #Maximum value in list x
    int nb=1000         #nb to calculate bcv
    ) nogil:

    '''
    Cython version of bcv
    '''

    cdef np.double_t h, lower, upper
    cdef density den
    if(n == 0):
        printf("'x' has length zero\n")
        return -1
    hmax = 1.144 * sqrt(r_var(x, n)) * (n**(-1.0/5.0)) * 4 #Why *4?
    lower = 0.1 * hmax
    upper = hmax
    den = VR_den_bin(n, nb, x, x_min, x_max)
    
    # Optimize
    cdef np.double_t x1 = lower, x2 = upper, xatol = 0.1*lower
    cdef int maxfun = 500, num
    cdef np.double_t sqrt_eps = sqrt(2.2e-16), golden_mean = 0.5 * (3.0 - sqrt(5.0))
    cdef np.double_t a, b, nfc, xf, fulc, 
    cdef np.double_t rat, e, fx, ffulc, fnfc, xm, tol1, tol2, r, p, q, golden, si, fu

    a = x1; b = x2
    fulc = a + golden_mean * (b - a)
    nfc, xf = fulc, fulc
    rat = e = 0.0
    h = xf
    
    fx = VR_bcv_bin(n, nb, den, h)
    num = 1

    ffulc = fnfc = fx
    xm = 0.5 * (a + b)
    tol1 = sqrt_eps * fabs(xf) + xatol / 3.0
    tol2 = 2.0 * tol1

    while (fabs(xf - xm) > (tol2 - 0.5 * (b - a))):
        golden = 1.0
        # Check for parabolic fit
        if fabs(e) > tol1:
            golden = 0.0
            r = (xf - nfc) * (fx - ffulc)
            q = (xf - fulc) * (fx - fnfc)
            p = (xf - fulc) * q - (xf - nfc) * r
            q = 2.0 * (q - r)
            if q > 0.0:
                p = -p
            q = fabs(q)
            r = e
            e = rat

            # Check for acceptability of parabola
            if ((fabs(p) < fabs(0.5*q*r)) and (p > q*(a - xf)) and
                    (p < q * (b - xf))):
                rat = (p + 0.0) / q
                h = xf + rat

                if ((h - a) < tol2) or ((b - h) < tol2):
                    si = sign(xm - xf) + ((xm - xf) == 0.0)
                    rat = tol1 * si
            else:      # do a golden section step
                golden = 1.0

        if golden:  # Do a golden-section step
            if xf >= xm:
                e = a - xf
            else:
                e = b - xf
            rat = golden_mean*e

        si = sign(rat) + (rat == 0.0)
        h = xf + si * fmax(fabs(rat), tol1)
        fu = VR_bcv_bin(n, nb, den, h)
        num += 1

        if fu <= fx:
            if h >= xf:
                a = xf
            else:
                b = xf
            fulc, ffulc = nfc, fnfc
            nfc, fnfc = xf, fx
            xf, fx = h, fu
        else:
            if h < xf:
                a = h
            else:
                b = h
            if (fu <= fnfc) or (nfc == xf):
                fulc, ffulc = nfc, fnfc
                nfc, fnfc = h, fu
            elif (fu <= ffulc) or (fulc == xf) or (fulc == nfc):
                fulc, ffulc = h, fu

        xm = 0.5 * (a + b)
        tol1 = sqrt_eps * fabs(xf) + xatol / 3.0
        tol2 = 2.0 * tol1

        if num >= maxfun:
            break
    h = xf
    free(den.cnt)

    return h

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cdef np.double_t local_pearsonr(
    np.double_t[:] x,   #list x
    np.int_t[:] y,      #list y
    int n               #size of list above
    ) nogil:

    '''
    Local function to calculater perason correlation without gil 
    '''

    cdef np.double_t mx, my, r_num = 0, r_den, r
    cdef np.double_t sum_x = 0, sum_y = 0
    cdef np.double_t xm, ym
    cdef int i
    mx = x[0]
    my = y[0]
    for i in range(1, n):
        mx += x[i]
        my += y[i]
    mx = mx / n
    my = my / n
    
    for i in range(n):
        xm = x[i] - mx
        ym = y[i] - my
        r_num += xm * ym
        sum_x += xm ** 2
        sum_y += ym ** 2
    r_den = sqrt(sum_x * sum_y)
    r = r_num / r_den
    return r

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cpdef int calculategridsize(
    np.double_t[:] y,           #Target
    np.int_t[:] x,              #Features to calculate correlation
    int k,                      #K value for kernel
    int size,                   #Number of sample
    np.double_t bandwidth,      #Value of bandwidth
    np.double_t bandwidth_mult, #Multiplier of bandwidth
    np.double_t bandwidth_adj,  #Adjust of bandwidth
    int neighborhood,           #Multiplication for bandwidth
    np.double_t deltay,
    np.double_t miny,
    np.double_t jitter=1E-10    #jitter to be added
    ) nogil:


    """ Calculate IC between continuous 'y' and binary 'x' """
    cdef np.double_t  cor, y_bandwidth, sigma_y, deltak
    cdef np.double_t  term_2_pi_sigma_y, integral, mutual_information
    cdef np.double_t  pX_0, pX_1, pysum, py0sum, py1sum
    cdef int y_d, i, j, sumx, grid
    cdef np.double_t* normy
    cdef np.double_t* kernal
    cdef np.double_t* p_y_total
    cdef np.double_t* p_y_1
    cdef np.double_t* p_y_0

    sumx = 0
    i = 0
    for i in range(size):
        sumx += <int>x[i]
    
    # Calculate bandwidth
    cor = local_pearsonr(y, x, size)
    
    y_bandwidth = bandwidth * (bandwidth_mult * (1 + (bandwidth_adj) * fabs(cor)))

    deltak = y_bandwidth * neighborhood
    grid = <int>(round(deltay/deltak*k))
    
    return grid

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef np.int_t[:] calcGrids(
    np.double_t[:] y_in,        #Target
    np.int_t[:,:] xs_in,        #List of features
    int k,                      #K value for kernel
    int n,                      #Number of result
    int size,                   #Number of sample
    np.double_t bandwidth,      #Value of bandwidth
    np.double_t bandwidth_mult, #Multiplier of bandwidth
    np.double_t bandwidth_adj,  #Adjust of bandwidth
    direction,                  #Direction that feature should match with target
    int grid,                   #Size of grid
    int thread_number,          #Number of thread
    int neighborhood
    ):

    '''
    Function to find feature with best IC among all features
    '''

    cdef np.int_t[:] res = np.zeros(n,dtype=int)
    cdef np.double_t[:] y = np.asarray(y_in)
    cdef np.int_t[:,:] x = np.asarray(xs_in)

    cdef int i
    cdef int maxseedid = 0
    cdef np.double_t maxCIC = 0.0
    cdef np.double_t miny = findmin(y, size)
    cdef np.double_t deltay = findmax(y, size) - findmin(y, size)
    
    for i in prange(n, nogil=True,num_threads=thread_number):
        res[i] = calculategridsize(y, x[i,:],k,size, bandwidth,bandwidth_mult,
                                                     bandwidth_adj, neighborhood, deltay, miny)
    
    return res

def drawTarget(
    comb    #Combined dataframe 
    ):
    
    '''
    Function to plot target
    '''

    f, ax = plt.subplots(figsize=(10, 1))
    ax = sns.heatmap(comb.iloc[[0]].to_numpy(), cmap='bwr', annot=False, yticklabels=False,
                     xticklabels=False, cbar=False, center=comb.iloc[0].mean())
    return f

def drawFeature(
    comb,           #Combined dataframe
    featurecmap,    #Cmap to plot feature
    seed_name=None, #Seed name to plot
    seedID=None     #Seed id to plot
    ):

    '''
    Function to plot feature
    '''

    if seed_name != None:
        f, ax = plt.subplots(figsize=(10, 1))
        ax = sns.heatmap(comb.loc[[seed_name]].to_numpy(), cmap=featurecmap, annot=False,
                         yticklabels=False, xticklabels=False, cbar=False)
        return f
    else:
        f, ax = plt.subplots(figsize=(10, 1))
        ax = sns.heatmap(comb.iloc[[seedID]].to_numpy(), cmap=featurecmap, annot=False,
                         yticklabels=False, xticklabels=False, cbar=False)
        return f
    
def drawSeed(
    seed,       #Seed list
    seedcmap    #Cmap to draw seed
    ):

    '''
    Function to draw seed
    '''

    f, ax = plt.subplots(figsize=(10, 1))
    ax = sns.heatmap([seed], cmap=seedcmap, annot=False, yticklabels=False, 
                     xticklabels=False, cbar=False)
    return f

def seedCombine(
    currentseed,    #Current seed
    newseed         #New seed to be added
    ):

    '''
    Function to combine two seed
    '''

    seed = []
    for i in range(len(currentseed)):
        if currentseed[i] == 1 or newseed[i] == 1:
            seed.append(1)
        else:
            seed.append(0)

    return seed

cpdef topmatch(
    comb,           #Combined dataframe
    currentseed,    #Current seed to run IC/CIC
    y,              #Target value
    grid,           #Size of grid
    k,              #K value for kernel
    size,           #Number of sample
    bandwidth,      #Value of bandwidth
    bandwidth_mult, #Multiplier of bandwidth
    bandwidth_adj,  #Adjust of bandwidth
    thread_number,  #Number of thread to run
    direction,      #Direction that feature should match with target
    num_top,        #Number of top feature
    prefix,         #Prefix for generated files
    figure_format,  #Format for figures
    ifSeed,         #Indicate if seed exist
    if_pval,        #Indicate if p-values are calculated
    if_bootstrap,   #Indicate if variances are calculated
    if_cluster,     #Indicate if features are clustered
    collapse,       #Indicate if same features are collapsed together for intermediate files
    seedcmap,       #Cmap for seed heatmap
    featurecmap,    #Cmap for feature heatmap
    nitr,           #Itration number for this figure
    out_folder,     #Folder name to put results
    seed_name,      #List of seed names
    target_name,    #Name of target
    gmt,
    locusdic,
    neighborhood
    ):

    '''
    Function to create intermediate file for each run to report top features in each run
    '''

    CICs = []
    pVals = []
    bootstraps = []
    cdef np.double_t miny = findmin(y, size)
    cdef np.double_t deltay = findmax(y, size) - findmin(y, size)
    
    if ifSeed == True:
        cythonseed = np.asarray(currentseed).astype(int)
        rank = topMatches(comb=comb, y=y, grid=grid, k=k, size=size, bandwidth=bandwidth,
                          bandwidth_mult=bandwidth_mult, bandwidth_adj=bandwidth_adj,
                          thread_number=thread_number, neighborhood = neighborhood,seed=cythonseed)

        IC = binaryInformationCoefficient_cython(y, cythonseed, k, size, bandwidth, bandwidth_mult,
                                                     bandwidth_adj, neighborhood, deltay, miny)

        CICs.append(IC)
        if if_bootstrap == True:
            bootstrap = calcBootstrapIC(subcomb = pd.DataFrame([comb.iloc[0].tolist(),currentseed]),
                                        size = size, bandwidth = bandwidth, k = k, 
                                        bandwidth_mult = bandwidth_mult, 
                                        bandwidth_adj = bandwidth_adj, grid = grid, 
                                        thread_number = thread_number, IC = IC, neighborhood = neighborhood)
            bootstraps.append(bootstrap)

        if if_pval == True:
            pVal = calcIndivisualPvalIC(subcomb = pd.DataFrame([comb.iloc[0].tolist(),currentseed]),
                                        size = size, bandwidth = bandwidth, k = k, 
                                        bandwidth_mult = bandwidth_mult, 
                                        bandwidth_adj = bandwidth_adj, grid = grid, 
                                        thread_number =thread_number, IC = IC, neighborhood = neighborhood)
            pVals.append(pVal)

    else:
        currentseed = [0]*len(comb.columns)
        cythonseed = np.asarray(currentseed).astype(int)
        rank = topMatches(comb=comb,y = y,grid=grid,k=k,size = size,bandwidth = bandwidth,
                          bandwidth_mult=bandwidth_mult,bandwidth_adj=bandwidth_adj,
                          thread_number=thread_number, neighborhood = neighborhood)
        CICs.append('')
        if if_bootstrap == True:
            bootstraps.append('')
        if if_pval == True:
            pVals.append('')

    rankdf = pd.DataFrame(rank)
    if direction == 'pos':
        rankdf.sort_values(by=1, ascending=False, inplace=True)
        rank = rankdf.to_numpy()
    else:
        rankdf.sort_values(by=1, ascending=True, inplace=True)
        rank = rankdf.to_numpy()

        
    pilesize = []
    if collapse == True:
        pileres,ICs = pileRank(rank,num_top,prefix,out_folder,nitr)
        pilename = []

        for i in range(len(pileres)):
            pilename.append(pileres[i][0][0])
            pilesize.append(len(pileres[i]))
        rankfeature = pd.concat([comb.iloc[[0]],comb.loc[pilename]])
        CICs = CICs + list(ICs)
    else:
        rankfeature = pd.concat([comb.iloc[[0]],comb.loc[rank[:,0][:num_top]]])
        CICs = CICs + list(rank[:,1][:num_top])
    
    if ifSeed == True:
        if if_pval==True:
            results = prepPvalCIC(comb = comb, cythonseed = cythonseed, size = size, 
                                  bandwidth = bandwidth, k = k, bandwidth_mult = bandwidth_mult, 
                                  bandwidth_adj = bandwidth_adj, grid = grid, 
                                  thread_number =thread_number, neighborhood = neighborhood)
            for i in range(1,len(rankfeature.index.tolist())):
                if i != 1 and CICs[i] == CICs[i-1]:
                    pVals.append(pVals[-1])
                else:
                    #print('CIC:'+str(CICs[i]))
                    pVals.append(calcPvalCIC(results = results, IC = CICs[i], direction = direction))

        if if_bootstrap==True:
            for i in range(1,len(rankfeature.index.tolist())):
                if i != 1 and CICs[i] == CICs[i-1]:
                    bootstraps.append(bootstraps[-1])
                else:
                    bootstrap = calcBootstrap(subcomb = rankfeature.iloc[[0,i],],  
                                              cythonseed = cythonseed, size = size, 
                                              bandwidth = bandwidth, k = k, 
                                              bandwidth_mult = bandwidth_mult, 
                                              bandwidth_adj = bandwidth_adj, grid = grid,
                                              thread_number = thread_number, IC = IC, 
                                              neighborhood = neighborhood)
                    bootstraps.append(bootstrap)
    else:
        if if_pval==True:
            results = prepPvalIC(comb = comb, size = size, 
                                 bandwidth = bandwidth, k = k, bandwidth_mult = bandwidth_mult, 
                                 bandwidth_adj = bandwidth_adj, grid = grid,
                                 thread_number = thread_number, neighborhood = neighborhood)
            for i in range(1,len(rankfeature.index.tolist())):
                if i != 1 and CICs[i] == CICs[i-1]:
                    pVals.append(pVals[-1])
                else:
                    #print('CIC:'+str(CICs[i]))
                    pVals.append(calcPvalIC(results = results, IC = CICs[i], direction = direction))

        if if_bootstrap==True:
            for i in range(1,len(rankfeature.index.tolist())):
                if i != 1 and CICs[i] == CICs[i-1]:
                    bootstraps.append(bootstraps[-1])
                else:
                    bootstrap = calcBootstrapIC(subcomb = rankfeature.iloc[[0,i],], size = size, 
                                                bandwidth = bandwidth, k = k, 
                                                bandwidth_mult = bandwidth_mult, 
                                                bandwidth_adj = bandwidth_adj, 
                                                grid = grid, thread_number = thread_number, 
                                                IC = CICs[i], neighborhood = neighborhood)
                    bootstraps.append(bootstrap)

    savetopfig(plotcomb = rankfeature.copy(), CICs = CICs, cythonseed = cythonseed, 
               seedcmap = seedcmap, featurecmap = featurecmap, prefix = prefix, 
               figure_format = figure_format, pVals = pVals, num_top = num_top, 
               bootstraps = bootstraps.copy(), pilesize = pilesize, nitr = nitr, gmt = gmt,
               out_folder = out_folder, target_name = target_name, seed_name = seed_name,
               locusdic = locusdic)

    if if_cluster == True and collapse == False:
        plotCluster(plotcomb = rankfeature.copy(), CICs = CICs, size = size, 
                    cythonseed = cythonseed, seedcmap = seedcmap, featurecmap = featurecmap,
                    prefix = prefix, figure_format = figure_format, pVals = pVals, 
                    num_top = num_top, bootstraps = bootstraps, nitr = nitr, gmt = gmt,
                    out_folder = out_folder, target_name = target_name, seed_name = seed_name,
                    locusdic = locusdic)

    return comb,rank,rankfeature

@cython.boundscheck(False)
@cython.cdivision(True)
cpdef runCheckGrid(target_file='no', # gct file for target(continuous or binary)
                  feature_file='no', # gct file for features(binary)
                  seed_file='no', # file for seed, if not provided, feature file is used directly 
                  target_df='no', # dataframe for target
                  feature_df='no', # dataframe for features
                  seed_df='no', # dataframe for seed
                  prefix='test', # prefix for result files 
                  seed_name=None, # names for seed, should be a list of string indicating the name of seed
                  grid=34, # number of grid, default is 34
                  target_name=0, # name/index of target in target file. can be int n for nth row, or string s for row with index s
                  k=5, # size of kernel for k standard deviation away
                  bandwidth_mult=0.65, # multiplication for bandwidth
                  bandwidth_adj=-0.95, # adjustion value for bandwidth
                  direction='pos', # direction that feature should match with target
                  mode='single', # indicate if multiple parameter set is passes. if True, then prefix, k, grid, bandwidth_mult, and bandwidth_adj has to be a list
                  num_top=30, # number of top matches shown in intermediate file
                  low_threshold=3, # lowest threshold that feature with less than this value occurence will be removed
                  high_threshold=100, # highest threshold that feature with less than this value occurence will be removed
                  collapse=True, # indicate if same features are collapsed together for intermediate files
                  normalize='zerobase', # normalize method for target
                  gene_locus='None', # gene_locus file indicating gene name and location of that gene
                  verbose=1, # verbose level(if 0, no report)
                  max_iteration=-1, # maximum of iteration for best CIC discovery, automatic detection by IC value if -1 
                  thread_number=1, # number of core used for parallel computing.
                  figure_format='pdf', # format for result figure
                  subset='no', # if list of string passed, only columns in this list is picked for calculation
                  if_pval=True, # if True, p-values are calculated for all result
                  if_bootstrap=True, # if True, variance by bootstrap is calculated for all result
                  if_cluster=False, # if True, features in intermediate files are clustered with NMF 
                  if_intermediate=False, # if True, intermediate result with top CIC value features are reported
                  out_folder='.', # folder to put output files inside
                  gene_set=None,
                  gene_separator='_',
                  gmt_file = None,
                  alpha = 1,
                  neighborhood = 4
                 ):
    
    report = ''

    if verbose != 0:
        print(prefix+' start!')
    report = report + 'Run name: ' + str(prefix) + '\n'
    
    # if / is not at the end of output folder, add it.
    if out_folder[-1] != '/':
        out_folder = out_folder + '/'
    
    # if output folder does not exist, make it
    if not os.path.exists(out_folder):
        os.makedirs(out_folder)
    
    
    start = time.time()
    if (isinstance(target_df,str) or isinstance(feature_df,str)) and \
       (target_file == 'no' or feature_file == 'no') :
        raise Exception("Please put input!")
    if mp.cpu_count() <thread_number:
        if verbose != 0:
            print('Indicated Core Number exceed maximum available core. All core is used')
        thread_number = mp.cpu_count()
    if thread_number == -1:
        if verbose != 0:
            print('All core Used')
        thread_number = mp.cpu_count()
    report = report + 'Number of thread used: ' + str(thread_number) + '\n'
    
    if seed_file == 'no':
        seed_file = feature_file
    if isinstance(feature_df,str):
        comb,locusdic = readInput(target_file = target_file, feature_file = feature_file, 
                         seed_file = seed_file, seed_name = seed_name, target_name = target_name,
                         direction = direction, gene_locus = gene_locus, normalize = normalize,
                         low_threshold = low_threshold, high_threshold = high_threshold,
                         grid = grid, subset = subset, gene_set = gene_set, 
                         gene_separator = gene_separator, thread_number = thread_number)
    else:
        comb = pd.concat([target_df,seed_df,feature_df], join='inner')
        rmrow=[]
        for i in comb.index.tolist()[2:]:
            if sum(comb.loc[i].tolist()) <= low_threshold or \
               sum(comb.loc[i].tolist()) >= high_threshold:
                rmrow.append(i)
        comb = comb.drop(index=rmrow)
        origcomb = comb.copy()
        seed_name = ['firstseed']
        locusdic = None
    if verbose != 0:
        print('Time used to read input: %s second(s)'%(float(time.time() - start)))
    
    if gmt_file != None:
        gmt = pd.read_csv(gmt_file,sep='\t',header=None,index_col=0)
        gmt = gmt.drop(columns=gmt.columns[0])
        gmt.index = gmt.index.str.replace("-", "_")
        if (len(list(set(comb.index.tolist())&set(gmt.index.tolist()))) == 0) and verbose != 0:
            print('Names in gmt file is not matching with input file. Setting gmt to None.')
            gmt = None
    else:
        gmt = None

    if mode == 'multi':
        for ind in range(len(prefix)):
            gridlist = REVEALERInner(prefix = prefix[ind], comb = comb, 
                                                           grid = grid[ind], k = k[ind], 
                                                           bandwidth_mult = bandwidth_mult[ind], 
                                                           bandwidth_adj = bandwidth_adj[ind], 
                                                           direction = direction, num_top = num_top,
                                                           verbose = verbose, 
                                                           max_iteration = max_iteration, 
                                                           collapse = collapse,
                                                           thread_number =thread_number, 
                                                           figure_format = figure_format, 
                                                           if_pval = if_pval, seed_name = seed_name,
                                                           if_bootstrap = if_bootstrap, 
                                                           if_cluster = if_cluster, 
                                                           if_intermediate = if_intermediate, 
                                                           out_folder = out_folder, 
                                                           normalize = normalize, 
                                                           target_name = target_name,
                                                           report = report,
                                                           gmt = gmt,
                                                           locusdic = locusdic,
                                                           alpha = alpha,
                                                           neighborhood = neighborhood)
        
    elif mode == 'single':
        gridlist = REVEALERInner(prefix = prefix, comb = comb, grid = grid, 
                                                       k = k, bandwidth_mult = bandwidth_mult, 
                                                       normalize = normalize, 
                                                       bandwidth_adj = bandwidth_adj, 
                                                       direction = direction, num_top = num_top, 
                                                       verbose = verbose, 
                                                       max_iteration = max_iteration,
                                                       thread_number =thread_number, 
                                                       figure_format = figure_format, 
                                                       if_pval = if_pval, seed_name = seed_name, 
                                                       if_bootstrap = if_bootstrap, 
                                                       if_cluster = if_cluster, 
                                                       if_intermediate = if_intermediate, 
                                                       out_folder = out_folder,
                                                       collapse = collapse, 
                                                       target_name = target_name,
                                                       report = report,
                                                       gmt = gmt,
                                                       locusdic = locusdic,
                                                       alpha = alpha,
                                                       neighborhood = neighborhood)
        
    return gridlist

@cython.boundscheck(False)
@cython.cdivision(True)   
cpdef REVEALERInner(
    prefix,             #Prefix for generated files
    comb,               #Combined dataframe
    grid,               #Size of grid
    k,                  #K value for kernel
    bandwidth_mult,     #Multiplier of bandwidth
    bandwidth_adj,      #Adjust of bandwidth
    direction,          #Direction that feature should match with target
    num_top,            #Number of top feature
    verbose,            #Verbose state
    max_iteration,      #Maximum of iteration for best CIC discovery
    collapse,           #Indicate if same features are collapsed together for intermediate files
    thread_number,      #Number of thread to run
    figure_format,      #Format for figures
    if_pval,            #Indicate if p-values are calculated
    if_bootstrap,       #Indicate if variances are calculated
    if_cluster,         #Indicate if features are clustered
    if_intermediate,    #Indicate if intermediate files are generated
    out_folder,         #Folder name to put results
    seed_name,          #List of seed names
    normalize,          #Normalization method
    target_name,        #Name of target
    report,
    gmt,
    locusdic,
    alpha,
    neighborhood
    ):

    seedcmap = clr.LinearSegmentedColormap.from_list('custom greys', [(.9,.9,.9),(0.5,0.5,0.5)], 
                                                     N=256)
    featurecmap = clr.LinearSegmentedColormap.from_list('custom greys', 
                                                        [(<double>(176)/<double>(255),
                                                        <double>(196)/<double>(255),
                                                        <double>(222)/<double>(255)),
                                                        (0,0,<double>(139)/<double>(255))], N=256)


    newpheno = []
    for i in comb.iloc[0].tolist():
        newpheno.append(np.sign(i)*(abs(i)**alpha))
    comb.iloc[0] = newpheno
    
    if verbose != 0:
        if seed_name != None:
            print("Number of features that pass the threshold is: "+str(len(comb.index)-2))
            print("Number of sample is: " + str(len(comb.columns)))
        else:
            print("Number of features that pass the threshold is: "+str(len(comb.index)-1))
            print("Number of sample is: " + str(len(comb.columns)))

    if seed_name != None:
        report = report + 'Number of features passing threshold is: ' + str(len(comb.index)-2)+'\n'
    else:
        report = report + 'Number of features passing threshold is: ' + str(len(comb.index)-1)+'\n'
    report = report + "Number of sample is: " + str(len(comb.columns)) + '\n'

    start= time.time()
    cdef int size = len(comb.iloc[0])
    cdef np.double_t[:] y = np.ascontiguousarray(np.asarray(comb.iloc[0].tolist()))
    cdef np.double_t miny = findmin(y, size)
    cdef np.double_t deltay = findmax(y, size) - findmin(y, size)
    i = 0
    cdef np.double_t bandwidth = calc_bandwidth(y,size)                                    
                                                
    cdef np.int_t[:] cythonseed
    
    seedids = []
    CICs = []
    seedIC = []
    pVals = []
    bootstraps = []
    addseed = 0
    
    if grid == None:
        grid = len(comb.columns.tolist())+1
    if verbose != 0:
        print('bandwidth: '+str(bandwidth))

    report = report + 'Universal bandwidth is: ' + str(bandwidth) + '\n'

    savecomb = comb.copy()
    seedlists = []

    gridlist = np.array(calcGrids(y, np.array(comb.iloc[1:].values.astype(int).tolist()),
                             k, int(len(comb.index)-1), size, bandwidth,
                             bandwidth_mult, bandwidth_adj, direction,grid,thread_number, neighborhood))
    
    plt.hist(gridlist)
    plt.title('Grid Distribution of '+prefix+'when k = '+str(k))
    plt.savefig(out_folder + prefix+'_grid_k'+str(k)+'.'+
                figure_format, format=figure_format)
    plt.close()
    return gridlist
        

def seedcomball( 
    comb,       #Dataframe 
    seed_name   #List of seed name
    ):
    
    '''
    combine all seed with given seed names
    '''

    currentseed = comb.loc[seed_name[0]].tolist()
    if len(seed_name) == 1:
        return currentseed
    for subseed in seed_name[1:]:
        currentseed = seedCombine(currentseed,comb.loc[subseed].tolist())
    return currentseed

def readInput(
    target_file,    #Name of target file
    feature_file,   #Name of feature file
    seed_file,      #Name of seed file
    seed_name,      #List of seed name(s)
    target_name,    #Target name
    direction,      #Direction to sort target
    gene_locus,     #Name of gene locus file
    normalize,      #Name of normalization method
    low_threshold,  #Lower threshold to filter feature
    high_threshold, #Higher threshold to filter feature
    grid,           #Gird size
    subset,         #List to subset target file if no subset, indicated 'no'
    gene_set,
    gene_separator,
    thread_number
    ):

    '''
    Function to read input files to generate combined dataframe
    '''
    
    print('Start reading input files...')

    #Read target file with given target name or index, replace - with _ in column name
    if isinstance(target_name,int) == True:
        target = pd.read_csv(target_file,skiprows=[0,1],sep='\t',index_col=0)
        target = target.drop(columns=target.columns[0]).iloc[[target_name]]
        target.columns = target.columns.str.replace("-", "_")
    else:
        target = pd.read_csv(target_file,skiprows=[0,1],sep='\t',index_col=0)
        target = target.drop(columns=target.columns[0]).loc[[target_name]]
        target.columns = target.columns.str.replace("-", "_")

    #Read seed file and feature file and extract seed. If same, read only one
    if seed_file == feature_file:
        feature = pd.read_csv(feature_file,skiprows=[0,1],sep='\t',index_col=0)
        feature = feature.drop(columns=feature.columns[0])
        feature.columns = feature.columns.str.replace("-", "_")
        comb = pd.concat([target,feature],  join='inner')
    else:
        feature = pd.read_csv(feature_file,skiprows=[0,1],sep='\t',index_col=0)
        feature = feature.drop(columns=feature.columns[0])
        seed_df = pd.read_csv(seed_file,skiprows=[0,1],sep='\t',index_col=0)
        seed_df = seed_df.drop(columns=seed_df.columns[0]).loc[seed_name]
        feature.columns = feature.columns.str.replace("-", "_")
        seed_df.columns = seed_df.columns.str.replace("-", "_")
        comb = pd.concat([target,seed_df,feature],  join='inner')

    #If subset list is passed, extract subset only
    if subset != 'no':
        if(len(list(set(comb.columns.tolist())&set(subset)))) == 0:
            print('subset not exist! Please check your subset list.')
            sys.exit(1)
        comb = comb[list(set(comb.columns.tolist())&set(subset))]

    #Sort whole dataframe by target by direction
    if direction == 'neg':
        comb = comb.sort_values(by = comb.index[0], axis = 1)
    else:
        comb = comb.sort_values(by = comb.index[0], axis = 1,ascending=False)

    comb = comb.dropna(axis=1, how='all')
    comb = comb.dropna(axis=0, how='all')
    
    comb = comb.dropna(axis=1)
    comb = comb.dropna(axis=0)

    #remove row by low and high threshold
    if low_threshold < 1:
        low_threshold = len(comb.iloc[0].tolist()) * low_threshold
    if high_threshold < 1:
        high_threshold = len(comb.iloc[0].tolist()) * high_threshold
    rmrow = []
    for i in comb.index.tolist()[1:]:
        if (sum(comb.loc[i].tolist()) <= low_threshold or sum(comb.loc[i].tolist()) >= high_threshold) and ((seed_name == None) or ((seed_name != None) and (i not in seed_name))):
            rmrow.append(i)
    comb = comb.drop(index=rmrow)
    seedIndex = []
    
    #Save index of seed    
    if seed_name != None:
        for i in seed_name:
            seedIndex.append(comb.index.tolist().index(i))
    

    if gene_locus != 'None':
        locus = pd.read_csv(gene_locus,sep='\t')
        locus = locus.dropna(axis=0)
        locusdic = {}
        for name in comb.index.tolist()[1:]:
            genename = name.split(gene_separator)[0]
            if genename in locus['Name'].tolist():
                locusdic[name] = locus[locus['Name']==genename]['Location'].tolist()[0]
            else:
                locusdic[name] = ''
    else:
        locusdic = None

    #Set first seed by extracting specific index part
    if seed_name != None:
        newseed_names = []
        for i in seedIndex:
            newseed_names.append(comb.index.tolist()[i])
        currentseed = seedcomball(comb,newseed_names)
        comb = comb.drop(index=newseed_names)
        comb.loc['firstseed'] = currentseed
        if gene_set != None:
            inset_list = []
            for feature_name in comb.index.tolist()[2:]:
                inset_list.append(check_inset(gene_set,feature_name,gene_separator))
            inset_list = [True,True]+inset_list
            comb = comb.iloc[inset_list]
    else:
        if gene_set != None:
            inset_list = []
            for feature_name in comb.index.tolist()[1:]:
                inset_list.append(check_inset(gene_set,feature_name,gene_separator))
            inset_list = [True]+inset_list
            comb = comb.iloc[inset_list]
            

    print('Done reading input files')
    
    return comb,locusdic

def check_inset(gene_set,feature_name,gene_separator):
    if feature_name.split(gene_separator)[0] in gene_set:
        return True
    else:
        return False
